package com.bored.services.tools.uds 
{
	import com.bored.services.BoredServices;
	import com.bored.services.events.DataReceivedEvent;
	import com.bored.services.events.ObjectEvent;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.TimerEvent;
	import flash.net.SharedObject;
	import flash.utils.Dictionary;
	import flash.utils.Timer;
	
	/**
	 * Dispatched when the user has logged-in, and the client-side Flash SharedObject was updated from server-side data, noting that the 
	 * Flash SharedObject data is dirty.
	 * @eventType com.bored.services.tools.uds.DataUpdatedEvent.UDS_SYNC_EVENT
	 */
	[Event(name = "UDSSyncEvent", type = "com.bored.services.tools.uds.DataUpdatedEvent")]
	
	/**
	 * Dispatched when the user has logged-in, and the server's data at any key differs from 
	 * what is currently stored at that key.
	 * @eventType com.bored.services.tools.uds.DataUpdatedEvent.UDS_SERVER_UPDATE_EVENT
	 */
	[Event(name = "UDSServerUpdateEvent", type = "com.bored.services.tools.uds.DataUpdatedEvent")]
	
	/**
	 * This class is a sample implementation of the BoredServices UDS functions.
	 * 
	 * Use this class to automatically sync your SharedObject data with the Universal Data Storage available to logged-in Bored.com users.
	 * 
	 * @author bored.com
	 * 
	 * @see com.bored.services.BoredServices
	 * @see com.bored.services.BoredServices#getData()
	 * @see com.bored.services.BoredServices#setData()
	 */
	public class UniversalDataStorageObject extends EventDispatcher
	{
		/**
		 * The storage key (set via constructor) used for referencing the data within both the SharedObject and the Universal Data Object (if logged in).
		 */
		protected var _storageKey:String;
		
		/**
		 * The Client-side SharedObject using the given _storageKey to get/set all the data for this class.
		 * @see #_storageKey
		 */
		public var clientFSO:SharedObject;
		
		/**
		 * A dictionary of resolution functions stored by the keys to compare the data at.
		 * @see #buildResolveFuncs()
		 */
		protected var resolutionsDict:Dictionary;
		
		/**
		 * Construct a new UDS object for a specific key.  If the client is logged-in (or when the client logs-in), the FSO and UDS will be sync'd via the syncFSOWithUDS function.
		 * @param	a_storageKey:	key to the data.  This key is used as in SharedObject.getLocal(a_storageKey)
		 * 
		 * @see #syncFSOWithUDS()
		 * @see #setDataAt()
		 */
		public function UniversalDataStorageObject(a_storageKey:String) 
		{
			if (!a_storageKey)
			{
				throw new Error("UniversalDataStorageObject::UniversalDataStorageObject(): a_storageKey=" + a_storageKey+ ", construction argument 'a_storageKey' must NOT resolve to false.");
			}
			
			if (!BoredServices.servicesObj)
			{
				throw new Error("UniversalDataStorageObject::UniversalDataStorageObject(): BoredServices.servicesObj=" + BoredServices.servicesObj 
								+ ". Please make sure that [DocumentClass].setBoredServices() is called prior to your application's remaining construction.");
			}
			
			_storageKey = a_storageKey;
			clientFSO = SharedObject.getLocal(_storageKey);
			
			// build our resolution dictionary (this also holds local data.
			resolutionsDict = new Dictionary();
			
			// first, populate the local UDS data with whatever's in the FSO at this point (could be up-to-date).
			populateLocalDataWithFsoData();
			
			// run this protected function to build any non-default resolve functions, should this class be subclassed.
			buildResolveFuncs();
			
			if (BoredServices.isLoggedIn)
			{
				onLoggedIn();
			}
			else
			{
				BoredServices.addEventListener(ObjectEvent.LOGGED_IN_EVENT, onLoggedIn, false, 0, true);
			}
			
		}//end UniversalDataStorageObject() constructor.
		
		/**
		 * This is used as a placeholder to be overridden in subclasses (optional), in which this would be an excellent place to place code such as the following:
		 * <listing>
		 * 
		 * // Here, we want data stored at 'maxLevelFinished' to use the maximum value, independent of whether it was acquired from the client or server.
		 * setResolutionFuncFor("maxLevelFinished", function(a_serverSideData:*, a_clientSideData:*):*{ return Math.max(a_serverSideData, a_clientSideData); } );
		 * 
		 * </listing>
		 */
		protected function buildResolveFuncs():void
		{
			
		}//end buildResolveFuncs()
		
		/**
		 * 
		 * @param	a_key:				String key at which, should the server-data differ from what is currently stored, the resolution-function, a_resolutionFunc, is called.
		 * @param	a_resolutionFunc:	Used to resolve conflicts where the server's data for a specified key differs from that on the client. This function should have the following structure:
		 * <listing>
		 * function(a_serverSideData:&#42;, a_clientSideData:&#42;):&#42;
		 * {
		 * 		return RESOLVED_VALUE_TO_STORE_AT_THIS_KEY;
		 * }
		 * </listing>
		 */
		public function setResolutionFuncFor(a_key:String, a_resolutionFunc:Function):void
		{
			if (undefined == resolutionsDict[a_key])
			{
				resolutionsDict[a_key] = new Object();
			}
			
			resolutionsDict[a_key].resolve = a_resolutionFunc;
			
		}//end setResolutionFuncFor()
		
		/**
		 * @private
		 */
		private function populateLocalDataWithFsoData():void
		{
			for (var key:String in clientFSO.data)
			{
				this.setDataAt(key, clientFSO.data[key]);
			}
			
		}//end populateLocalDataWithFsoData()
		
		/**
		 * @private
		 */
		private function onLoggedIn(a_evt:Event = null):void
		{
			BoredServices.removeEventListener(ObjectEvent.LOGGED_IN_EVENT, onLoggedIn);
			
			// now that we're logged in, acquire the storage object at this key.
			BoredServices.addEventListener(DataReceivedEvent.SAVED_DATA_RECEIVED_EVT, onDataInitiallyAcquired, false, 0, true);
			
			// This will triggger the DataReceivedEvent.SAVED_DATA_RECEIVED_EVT event.
			BoredServices.getData(_storageKey);
			
		}//end onLoggedIn()
		
		/**
		 * Handler for acquiring the data at _storageKey, instigated by the getData() call made in onLoggedIn().
		 * 
		 * @private
		 * 
		 * @see #onLoggedIn()
		 * 
		 */
		private function onDataInitiallyAcquired(a_dataRecdEvt:DataReceivedEvent):void
		{
			if (_storageKey != a_dataRecdEvt.key)
			{
				// this isn't the same key for the data we're expecting...keep waiting.
				return;
			}
			
			// we've found the data for our key, stop waiting for new data.
			BoredServices.removeEventListener(DataReceivedEvent.SAVED_DATA_RECEIVED_EVT, onDataInitiallyAcquired);
			
			var dataFromServer:Object = a_dataRecdEvt.data;
			
			if (null == dataFromServer)
			{
				dataFromServer = new Object();
			}
			
			this.dispatchEvent(new DataUpdatedEvent(DataUpdatedEvent.UDS_SERVER_UPDATE_EVENT, null, null, dataFromServer));
			
			resolveWithDataFromServer(dataFromServer);
			
			updateServerData();
			
		}//end onDataInitiallyAcquired()
		
		/**
		 * This is where, if a resolution function was set for a particular key (via setResolutionFuncFor()) it would be called, if the current and server-side data for the key differ.
		 * If the data on the server for this key differs from the data on the client for this key, regardless if it was resolved via resolution-function or not, 
		 * an event is dispatched, DataUpdatedEvent.UDS_SYNC_EVENT.
		 * 
		 * @param	a_dataFromServer:	An object containing all of the key-value pairs for this user for this storage object.
		 * 
		 * @see #setResolutionFuncFor()
		 * @see com.bored.services.tools.uds.DataUpdatedEvent
		 */
		protected function resolveWithDataFromServer(a_dataFromServer:Object):void
		{
			var resolutionDictObj:Object;
			
			var updateEvents:Array = new Array();
			
			for(var key:String in a_dataFromServer)
			{
				var tmpFormerVal:*;
				
				// make the storage object for this key in case it didn't exist.
				buildNewStorageObjFor(key);
				
				resolutionDictObj = resolutionsDict[key];
				
				if (resolutionDictObj.hasOwnProperty("resolve"))
				{
					tmpFormerVal = resolutionDictObj.data;
					
					try
					{
						resolutionDictObj.data = resolutionDictObj.resolve(a_dataFromServer[key], tmpFormerVal);
					}
					catch (e:Error)
					{
						// note the error in resolving this data, and by default treat the server-side data as sacred.
						resolutionDictObj.data = a_dataFromServer[key];
					}
					
					if (tmpFormerVal !== resolutionDictObj.data)
					{
						// the data changed, note this with a sync event.
						updateEvents.push(new DataUpdatedEvent(DataUpdatedEvent.UDS_SYNC_EVENT, key, tmpFormerVal, resolutionDictObj.data));
					}
				}
				else
				{
					// no resolution function for this key.  Perhaps warn the user that they might be out-of-sync?
				}
				
				setDataAt(key, resolutionDictObj.data, false);
			}
			
			// dispatch events noting that a value has been updated.
			for (var i:int = 0; i < updateEvents.length; i++)
			{
				this.dispatchEvent(Event(updateEvents[i]));
			}
			
		}//end resolveWithDataFromServer()
		
		/**
		 * Acquire saved-data, here.
		 * 
		 * @param	a_key:	Key to the data object
		 * @return	Data stored at a_key.
		 */
		public function getDataAt(a_key:String):*
		{
			return clientFSO.data[a_key];
			
		}//end getDataAt()
		
		/**
		 * @private
		 * @param	a_key
		 */
		private function buildNewStorageObjFor(a_key:String):void
		{
			if (undefined == resolutionsDict[a_key])
			{
				// the object at this key hasn't been created yet, so, create it.
				resolutionsDict[a_key] = new Object();
				
				// set the default resolution function.
				resolutionsDict[a_key].resolve =	function(a_serverSideData:*):*
													{
														return a_serverSideData;
													};
				
			}
			
		}//end buildNewStorageObjFor()
		
		/**
		 * Set save-data, here. If no resolution function was set for this key, then server-data at this key <b>always</b> takes priority, and will overwrite client-stored data pre-login.
		 * 
		 * @param	a_key:		Key to save the data at.
		 * @param	a_value:	Value to save the data at a_key.
		 */
		public function setDataAt(a_key:String, a_value:*, a_doPushToServer:Boolean = true):void
		{
			var storageObj:Object;
			
			buildNewStorageObjFor(a_key);
			
			// storage object for this key already existed, get it.
			storageObj = resolutionsDict[a_key];
			
			// finally, set the data in the our local copies.
			storageObj.data = clientFSO.data[a_key] = a_value;
			clientFSO.setProperty(a_key, a_value);
			
			// and save the data server-side with clientFSO.data
			if (a_doPushToServer)
			{
				updateServerData();
			}
			
		}//end setDataAt()
		
		/**
		 * This is called internally to sync the UDS to clients' saved data. Generally there should be no need to call this from outside of an instance of this class.
		 */
		public final function updateServerData():void
		{
			if (BoredServices.isLoggedIn)
			{
				BoredServices.setData(_storageKey, clientFSO.data);
			}
			
		}//end updateServerData()
		
	}//end class UniversalDataStorageObject

}//end package com.bored.services.tools.uds
